
package provide tktocanvas 0.3
package require pdf 2.0
package require base64
#source pdf.tcl
#source transformations.tcl

# If this variable is set to true, then a tk window
# with drawn exported window is created (It differs from pdf output a little).
set ::DBG 0

namespace eval tktocanvas {
# Rotation angle.

variable rotAngle 0
# Rotation point.

variable rotPoint {0 0}

# Scale factors.

variable scaleX 1
variable scaleY 1

# Offset values.

variable offsetX 0
variable offsetY 0

# Current page stream, is cleared by pdfNextPage procedure.

variable pdfout ""

# Pdf file's data (width, height, ...)

variable paperdata ""

# A list of pages.

variable pages {}

variable pagewidth 612 
variable pageheight 792

variable marginleft 10
variable marginright 10
variable margintop 10
variable marginbottom 10

variable pagex 0
variable pagey 0

proc ValidatePdfOptions {opts} {

    foreach {k v} $opts {
        switch -- $k {
            "-pagewidth" - "-pageheight" {
                if {![regexp {^[[:digit:]]+(.[[:digit:]]+)?(([ic]?)|(mm))$} $v]} {
                    error "ERROR: '$v' is not valid value!"
                }
            }
            "-scalex" - "-scaley" - "-marginleft" - "-marginright" -\
            "-margintop" - "-marginbottom" - "-pagex" - "-pagey" {
                if {![regexp {[[:digit:]]+(.[[:digit:]]+)?} $v]} {
                    error "ERROR: scale factor should be a number."
                }
            }
            "-pageanchor" {
                if {![regexp {([nwse])|([sn][we])|([we][sn])} $v]} {
                    error "ERROR: pageanchor should be n, s, w or e."
                }
            }
        }
    }
}

proc SetMargins {} {
    
    variable marginleft
    variable marginright
    variable margintop
    variable marginbottom
    variable pagewidth
    variable pageheight
    
    #puts "0 0 \
            [expr {$pagewidth - $marginright - $marginleft}] \
            [expr {$pageheight - $marginbottom - $margintop}]"
    SetDrawableRegion 0 0 \
            [expr {$pagewidth - $marginright - $marginleft}] \
            [expr {$pageheight - $marginbottom - $margintop}]
}

proc EndMargins {} {
    
    UnsetRegion
}

proc setPdfOptions {width height} {

    # Scale option.

    if {([info exists ::tktocanvas::scalex] && $::tktocanvas::scalex != 1) \
            || ([info exists ::tktocanvas::scaley] \
            && $::tktocanvas::scaley != 1)} {
        setScaleFactors $::tktocanvas::scalex $::tktocanvas::scaley
    }

    # Rotation option.
   
    if {[info exists ::tktocanvas::rotatexy] && $::tktocanvas::rotatexy != 0} {
        
        set rotPoint [list [expr {int($width / 2)}] \
                [expr {int ($height / 2)}]]
        tktocanvas::setRotation $rotPoint 90

        set ::tktocanvas::pagey [expr {$::tktocanvas::pagey \
                - int(((0.5 * $height) - (0.5 * $width)))}]
        
        set offset [lindex [rotateXY {*}[DoScaling 0 $height]] 0]
        set ::tktocanvas::pagex [expr {$::tktocanvas::pagex - $offset - round($::tktocanvas::scaleY)}]
    }

    # Page offsets
    
    if {($::tktocanvas::pagex != 0) || ($::tktocanvas::pagey != 0)} {
        setOffsets $::tktocanvas::pagex $::tktocanvas::pagey
    }

    # Page anchor option. Warning it changes width and height variables.

    if {[info exists ::tktocanvas::pageanchor]} {
        set x 0
        set y 0
      
        set width [expr {$::tktocanvas::scaleX * $width}]
        set height [expr {$::tktocanvas::scaleY * $height}]

        if {$::tktocanvas::rotatexy != 0} {
            set tmp $width
            set width $height
            set height $tmp
        }

        set pgwidth [expr {$::tktocanvas::pagewidth \
                - $::tktocanvas::marginleft - $::tktocanvas::marginright}]

        set pgheight [expr {$::tktocanvas::pageheight \
                - $::tktocanvas::margintop - $::tktocanvas::marginbottom}]

        if {[string length $::tktocanvas::pageanchor] == 1} {
            switch -regexp $::tktocanvas::pageanchor {
                "^n$" {
                    set x [expr {($pgwidth / 2) - ($width / 2)}]
                }
                "^s$" {
                    set x [expr {($pgwidth / 2) - ($width / 2)}]
                    set y [expr {$pgheight - $height}]
                }
                "^e$" {
                    set x [expr {$pgwidth - $width}]
                    set y [expr {($pgheight / 2) - ($height / 2)}]
                }
                "^w$" {
                    set y [expr {($pgheight / 2) - ($height / 2)}]
                }
            }
        } else {
            if {[regexp "e" $::tktocanvas::pageanchor]} {
                set x [expr {$pgwidth - $width}]
            }
            if {[regexp "s" $::tktocanvas::pageanchor]} {
                set y [expr {$pgheight - $height}]
            }
        }

        set ::tktocanvas::pagex [expr {$::tktocanvas::pagex + $x}]
        set ::tktocanvas::pagey [expr {$::tktocanvas::pagey + $y}]
    }
}

proc RestartPdfOptions {} {
    
    variable rotAngle 0
    variable rotPoint {0 0}
    variable scaleX 1
    variable scaleY 1
    variable offsetX 0
    variable offsetY 0
    variable pages {}
    variable pagewidth 612 
    variable pageheight 792
    variable marginleft 10
    variable marginright 10
    variable margintop 10
    variable marginbottom 10
    variable pagex 0
    variable pagey 0
    
    variable pdfout ""
    variable paperdata ""
    variable pages {}
}

proc startFile {args} {

    variable paperdata

    ValidatePdfOptions $args

    update

    set fileName [lsearch $args "-file"]
    if {$fileName < 0} {
        error "ERROR: You must specify -file option."
    }

    set fileName [lindex $args $fileName+1]

    pdf::initPDF_Write $fileName
    set paperdata $args
    
    foreach {k v} $args {
        set k [string range $k 1 end] 
        
        if {[regexp {[imcp]$} $v]} {
            set v [UnitsToPixels $v] 
        }
        
        variable $k $v
    }

    SetMargins
}

proc nextPage {} {
    
    variable pages
    variable pdfout

    EndMargins
    lappend pages $pdfout
    set pdfout ""
    SetMargins
}

proc endFile {} {
    
    variable paperdata
    variable pdfout
    variable pages

    EndMargins
    lappend pages $pdfout
   
    eval pdf::docTree_Write -pagecount [llength $pages] $paperdata
    
    pdf::Directive_Write /PDF /Text /ImgageC
    
    # Write all pages.
    
    foreach Page $pages { 
        pdf::pageObj_Write $Page
    }
    
    pdf::finish_Write

    RestartPdfOptions
}

proc GetRotationMatrix {x y angle} {
   
    switch $angle {
        90 {
            set rotString "0 -1 1 0 [expr {$x - $y}] [expr {$x + $y}]"
        }
        180 {
            set rotString "-1 0 0 -1 [expr {2 * $x}] [expr {2 * $y}]"
        }
        270 {
            set rotString "0 1 -1 0 [expr {$x + $y}] [expr {-$x + $y}]"
        }
        default {
            set rotString "1 0 0 1 0 0"
        }
    }

    return $rotString
}

proc GetScaleMatrix {x y} {

    set e [expr {-($::tktocanvas::scaleX - 1) * $x}]
    set f [expr {-($::tktocanvas::scaleY - 1) * $y}]

    return "$::tktocanvas::scaleX 0 0 $::tktocanvas::scaleY $e $f"
}

proc UnitsToPixels {screenDistance} {
    set val [string range $screenDistance 0 end-1]
    
    if {[string equal $val ""]} {
        set val [string range $screenDistance 0 end]
    }

    set pixval [expr {round([winfo fpixels . 1p])}]
    
    switch -glob -- [string range $screenDistance end end] {
        c { return [expr {$val * 28.8 * $pixval}] }
        i { return [expr {$val * 72 * $pixval}]   }
        m { return [expr {$val * 288 * $pixval}]  }
        p { return $pixval} 
        default {return $val}
    }
}

#proc Screen2Points {x} {
#    return $x 
#    return [expr ($x * (1.0 / [winfo fpixels . 1p]))]
#}

#############################################################################
# Procedure: FormatCorners - Restate the coordinates for the bounding
#                            rectangle of a canvas item within the page
#                            framework.
# Inputs:
#   coords - Ordered pair of x-y coordinates.
# Output:
#   List of four corner coordinates.
#

proc FormatCorners {x1 y1 x2 y2} {

    if {$x1 <= $x2} {
        set xa $x1
        set xb $x2
    } else {
        set xa $x2
        set xb $x1
    }
    if {$y1 <= $y2} {
        set ya $y1
        set yb $y2
    } else {
        set ya $y2
        set yb $y1
    }
    
    return [list $xa $ya $xb $yb]
}

proc PdfRectangle {x y x1 y1 fill outline {linewidth 1}} {
 
    variable pdfout

    foreach {x y x1 y1} [TkToPdfCoords [list $x $y $x1 $y1]] break

    set width [expr {$x1 - $x}]
    set height [expr {$y1 - $y}]
    
    append pdfout [::pdf::pdfRectangle $x $y $width \
            $height $outline $fill $linewidth]
}

proc PdfArc {x y x1 y1 fill outline start extent style {linewidth 1}} {

    variable pdfout

    set coords [TkToPdfCoords "$x $y $x1 $y1"]
    
    append pdfout [::pdf::pdfArc $linewidth $outline $fill \
            {*}$coords $start $extent $style]
}

# TODO: arrow, arrowshape, splinesteps
proc PdfLine {coords fill {linewidth 1} {smooth "none"} \
        {arrow "none"} {arrowshape "none"}} {

    variable pdfout
    
    set coords [TkToPdfCoords $coords]

    switch $smooth {
        "raw" {
            append pdfout [pdf::pdfRawCurve $linewidth $fill "" $coords]
        }
        "bezier" - "true" - "1" {
            append pdfout [pdf::pdfBezier $linewidth $fill "" $coords]
        }
        default {
            append pdfout [pdf::pdfLine $linewidth $fill $coords]
        }
    }
}

# TODO: splinesteps
proc PdfPolygon {coords fill outline {linewidth 1} {smooth "none"}} {

    variable pdfout
    
    set coords [TkToPdfCoords $coords]

    switch $smooth {
        "raw" {
            append pdfout [pdf::pdfRawCurve $linewidth $outline $fill $coords]
        }
        "bezier" - "true" - "1" {
            append pdfout [pdf::pdfBezier $linewidth $outline $fill $coords]
        }
        default {
            append pdfout [pdf::pdfPolygon $linewidth $outline $fill $coords]
        }
    }
}

proc PdfOval {coords fill outline {linewidth 1}} {
    
    variable pdfout

    set coords [TkToPdfCoords $coords]

    append pdfout [pdf::pdfOval {*}$coords $outline $fill $linewidth]
}

proc PdfText {x y font color txt justify width height} {

    variable pdfout
    variable pageheight

    # Catch justify and multiple line text.
   
    set numOfLines 0
    for {set i 0} {$i < [string length $txt]} {incr i} {
        if {[string index $txt $i] == "\n"} {
            incr numOfLines
        }
    }

    set lineSpace [font metrics $font -linespace]
    foreach {fam size style slant} $font break
    
    if {$size == ""} {
        set size 10
    } elseif {$size < 0} {
        set size [expr {-$size}]
    } else {
        set size [expr round($size * [tk scaling])]
    }

    set fontName [pdf::getFontName [string tolower $fam-$style-$slant]]
    if {$fontName == ""} {
        pdf::fontObj_Write /Type1 \
                [string tolower $fam-$style-$slant] /Default ;#/MacRomanEncoding 
        set fontName [pdf::getFontName [string tolower $fam-$style-$slant]]
    }
    
    set txt [split $txt "\n"]

    while {$numOfLines >= 0} {
        switch $justify {
            "center" {
                set xPos [expr {$x + ($width / 2) - [font measure $font \
                        [lindex $txt $numOfLines]]/2}] 
            }
            "right" {
                set xPos [expr {int($x + $width - [font measure $font \
                        [lindex $txt $numOfLines]])}]
            }
            default {
                set xPos $x
            }
        }

        set yTmp $y
        foreach {xPos y} [TransformPointsList [list $xPos $y]] break
        foreach {xPos y} [TkToPdfCoords [list $xPos $y]] break

        set scaleMatrix "[GetScaleMatrix $xPos $y]"
        
        append pdfout "q\n$scaleMatrix cm\n"
        
        set rotationMatrix "[GetRotationMatrix \
                $xPos $y $::tktocanvas::rotAngle] Tm\n"

        append pdfout [pdf::pdfTransformatedText $fontName $size \
                $color $xPos $y [lindex $txt $numOfLines] \
                "$rotationMatrix"]

        append pdfout "Q\n" 
       
        set y $yTmp
        incr numOfLines -1 
        switch $::tktocanvas::rotAngle {
            180 - 270 {
                set y [expr {$y - $lineSpace}]
            }
            default { 
                set y [expr {$y - $lineSpace}]
            }
        }
    }
}

proc PdfImage {x y imgName width height data} {
   
    #puts "$x $y $imgName $width $height data" 
    
    variable pdfout
    
    set x [expr {$x - ($width / 2)}]
    set y [expr {$y + ($height / 2)}]
    
    foreach {x y} [TransformImage [list $x $y] $width $height] break
    foreach {x y} [TkToPdfCoords [list $x $y]] break

    set data [string map {# {}} $data]
    set data [string map {\{ {}} $data]
    set data [string map {\} {}} $data]
    
    set imageObjID [pdf::getImageName $imgName]

    set geom [list $width 0 0 $height $x $y]
  
    set rotMatrix [GetRotationMatrix $x $y $::tktocanvas::rotAngle]
    set scaleMatrix [GetScaleMatrix $x $y] 
    append pdfout "q\n $rotMatrix cm\n $scaleMatrix cm\n"
    
    if {[string equal "" $imageObjID]} {
        pdf::pdfImageObj  $imgName $width $height $data
        set imageObjID [pdf::getImageName $imgName]
    }
    
    append pdfout "\nq\n$geom cm\n $imageObjID Do\nQ\n"
    append pdfout "Q\n"
}

proc TkToPdfCoords {pts} {

    variable pageheight
    variable marginleft
    variable margintop    

    set res {}
    
    foreach {x y} $pts {
        set y [expr {$pageheight - $y}]
        set y [expr {$y - $margintop}]
        set x [expr {$x + $marginleft}]
        set res "$res $x $y"
    }
    
    return $res
}

proc TkToPdfTextCoords {x y width height font {anchor "center"}} {
   
    if {$anchor != "center"} {
        if {[regexp "n" $anchor]} {
            set y [expr {$y + ($height / 2)}]
        }
        if {[regexp "s" $anchor]} {
            set y [expr {$y - ($height / 2)}]
        }
        if {[regexp "e" $anchor]} {
            set x [expr {$x - ($width / 2)}]
        }
        if {[regexp "w" $anchor]} {
            set x [expr {$x + ($width / 2)}]
        }
    }

    set y [expr {$y + int(($height / 2) - [font metrics $font -descent])}]
    set x [expr {$x - ($width/2)}]

    return [list $x $y]
}

proc SetDrawableRegion {x y x1 y1} {
    
    variable pdfout
    variable pageheight
   
    foreach {x y x1 y1} [TkToPdfCoords [list $x $y $x1 $y1]] break
#puts "$x $y $x1 $y1"
    append pdfout "q\n"
    append pdfout "$x $y m\n $x $y1 l\n $x1 $y1 l\n $x1 $y l\n W \n n\nB\n"
}

proc UnsetRegion {} {

    variable pdfout
    append pdfout "Q\n"
}

proc SetStandardOptions {itemName} {
    
    variable pdfout
   
    upvar $itemName item
    array set i [dict get $item]
    
    if {($i(-class) == "bitmap") || ($i(-class) == "image") \
            || ($i(-class) == "text")} {
        return
    }

    set optsString ""

    # Dash pattern.
    
    set dashPattern [SetDashPattern $i(-dash) $i(-dashoffset) $i(-width)]
    append pdfout "\[[lindex $dashPattern 0]\] [lindex $dashPattern 1] d\n"

    # Join style.

    if {[info exists i(-joinstyle)]} {
        
        switch $i(-joinstyle) {
            "bevel" {
                append pdfout "2 j\n"
            }
            "round" {
                append pdfout "1 j\n"
            }
        }
    }

    # Cap style.
    
    if {[info exists i(-capstyle)]} {
        switch $i(-capstyle) {
            "projecting" {
                append pdfout "2 J\n"
            }
            "round" {
                append pdfout "1 J\n"
            }
        }
    }

    append pdfout $optsString
}

proc SetDashPattern {dash dashOffset width} {
    
    set dashPattern ""
    
    if {![string is integer [string index $dash 0]]} {
        set strlen [string length $dash]
        set dashVal ""

        for {set i 0} {$i < $strlen} {incr i} {
            set char [string index $dash $i]

            switch $char {
                " " {
                    lset dashVal 1 [expr {[lindex $dashVal 1] + 4 * $width}]
                }
                "." {
                    append dashPattern "$dashVal "
                    set dashVal "[expr {2 * $width}] [expr {4 * $width}]"
                }
                "-" {
                    append dashPattern "$dashVal "
                    set dashVal "[expr {6 * $width}] [expr {4 * $width}]"
                }
                "," {
                    append dashPattern "$dashVal "
                    set dashVal "[expr {4 * $width}] [expr {4 * $width}]"
                }
                "_" {
                    append dashPattern "$dashVal "
                    set dashVal "[expr {8 * $width}] [expr {4 * $width}]"
                }
            }
        }
        
        append dashPattern " $dashVal "
        set dashOffset [expr {$dashOffset * $width}]
    } else {
        foreach {d d1} $dash {
            if {$d1 != ""} {
                append dashPattern "[expr {$d + $width}] \
                        [expr {($d1 - $width) < 0 ? 0 : ($d - $width)}] "
            } else {
                append dashPattern "$d "
            }
        }
    }
    
    return [list $dashPattern $dashOffset]
}

##############################################################################
# proc DrawHorizontalBar --
#
#       This procedure draws a horizontal bevel along one side of an object.
#       It's look like this:
#                       ----------------
#                        -------------
#                         ----------
#
# Arguments:
#       x, y, x1, y1 - Coords of lines which are on the most right/left
#                      side of the lines.
#       width        - Number of lines.
#       side         - 1 - top bevel; -1 bottom bevel.
#       colors       - A list with dark and shadow (in this order) colors
#                      of the bevel.
#       relief       - This parameters exchanges shadow colors depending
#                      on type of relief.
#       wrapoff      - 0 - means there is no 3d effect (lines make a 
#                      rectangle); 1 - there is left side 3d effect;
#                      2 - there is left side 3d effect; 3 - left and right
#                      3d effect. 
# Results:
#       A horizontal bevel is drawn.

proc DrawHorizontalBar {x y x1 y1 width side colors relief wrapoff} {

    if {$side < 0} {
        set y $y1
        incr x1
    }
    
    if {[expr {$wrapoff & 1}]} { 
        incr x
        incr x1
    } else {
        incr x1
    }
    
    set color [expr {$side > 0 ? [lindex $colors 0] : [lindex $colors 1]}]
    set hBorder [expr {$width / 2}]

    for {set d 0} {$d < $width} {incr d} {
       
        if {($d == $hBorder)&&(($relief == "groove")||($relief == "ridge"))} {
            set color [expr {$side < 0 \
                    ? [lindex $colors 0] : [lindex $colors 1]}]
            if {!$wrapoff} {
                if {$side == -1} {
                    if {[expr {$wrapoff & 2}]} {
                        incr x1 -$hBorder
                    }
                } else {
                    if {[expr {$wrapoff & 1}]} {
                        incr x $hBorder
                    }
                }
            }
        }
       
        if {$::DBG} {
            .t.c create line $x $y $x1 $y -fill $color
        }

        PdfLine [list [expr {$x - 1}] $y [expr {$x1 - 1}] $y] $color

        if {$side > 0} {
            incr y 1
        } else {
            incr y -1
        }
        
        if {$side == 1} {
            if {[expr {$wrapoff & 2}]} {
                incr x1 -1
            }
            if {[expr {$wrapoff & 3}] && $wrapoff != 2} {
                incr x 1
            }
        } elseif {$side == -1} {
            if {[expr {$wrapoff & 1}]} {
                incr x 1
            }
            if {[expr {$wrapoff & 3}] && $wrapoff != 1} {
                incr x1 -1
            }
        }
    }
}

##############################################################################
# proc DrawVerticalBar --
#
#       This procedure draws a vertical bevel along one side of an object.
#       It's look like this:
#                       |||
#                       |||
#                       |||
#
# Arguments:
#       x, y, x1, y1 - Coords of lines which are on the most right/left
#                      side of the lines.
#       width        - Number of lines.
#       side         - 1 - left bevel; -1 right bevel.
#       colors       - A list with dark and shadow (in this order) colors
#                      of the bevel.
#       relief       - This parameters exchanges shadow colors depending
#                      on type of relief.
# Results:
#       A vertical bevel is drawn.

proc DrawVerticalBar {x y x1 y1 width side colors relief} {
    
    if {$side < 0} {
        set x $x1
    }
 
    set color [expr {$side > 0 ? [lindex $colors 0] : [lindex $colors 1]}]
    set hBorder [expr {$width / 2}]

    for {set d 0} {$d < $width} {incr d} {
        if {($d == $hBorder)&&(($relief == "groove")||($relief == "ridge"))} {
            set color [expr {$side < 0 \
                    ? [lindex $colors 0] : [lindex $colors 1]}]
            if {$side == -1} {	
                incr y1 -$hBorder
            } else {
                incr y $hBorder
            }
        }
        
        if {$::DBG} {
            .t.c create line $x $y $x $y1 -fill $color
        }

        PdfLine [list $x $y $x $y1] $color 
       
        if {$side > 0} {
            incr x 1
        } else {
            incr x -1
        }
    }
}

#############################################################################
# proc Draw3DBorder --
#
#       Draws a 3d border around the widget.
#
# Arguments:
#       x            -
#       y            -
#       x1           -
#       y1           -
#       relief       -
#       borderwidth  -
#       shadowColors -
#
# Results:
#       A 3d border around the widget is drawn.

proc Draw3DBorder {x y x1 y1 relief borderwidth shadowColors} {

    # Exchange 3d colors when drawing rotated 3dborder.

    if {($::tktocanvas::rotAngle == 90) || ($::tktocanvas::rotAngle == 180)} {
        set tmp [lindex $shadowColors 0]
        lset shadowColors 0 [lindex $shadowColors 1]
        lset shadowColors 1 $tmp
    }

    DrawVerticalBar $x $y $x1 $y1 $borderwidth 1 $shadowColors $relief
    
    DrawVerticalBar $x $y $x1 $y1 $borderwidth -1 $shadowColors $relief


    if {($::tktocanvas::rotAngle == 90) || ($::tktocanvas::rotAngle == 270)} {
        set tmp [lindex $shadowColors 0]
        lset shadowColors 0 [lindex $shadowColors 1]
        lset shadowColors 1 $tmp
    }

    DrawHorizontalBar $x $y $x1 $y1 $borderwidth -1 $shadowColors $relief 3
    
    DrawHorizontalBar $x $y $x1 $y1 $borderwidth 1 $shadowColors $relief 3

}

proc DrawHighlightthickness {config} {

    array set c [dict get $config]

    for {set h $c(-highlightthickness)} {$h != 0} {incr h -1} {
        PdfRectangle [expr {$c(-x) - $h}] [expr {$c(-y) - $h}] \
            [expr {$c(-x1) + $h}] [expr {$c(-y1) + $h}] \
            "" $c(-highlightbackground)

        if {$::DBG} {
            .t.c create rectangle [expr {$c(-x) - $h}] [expr {$c(-y) - $h}]\
                [expr {$c(-x1) + $h}] [expr {$c(-y1) + $h}] \
                -outline $c(-highlightbackground)
        }
    }
}

# draws indicator for checkbutton/radiobutton
proc BtnDrawIndicator {config} {

    array set c [dict get $config]

    if {$c(-relief) == "sunken"} {
        set tmp $c(-lightshadow)
        set c(-lightshadow) $c(-darkshadow)
        set c(-darkshadow) $tmp
    }

    if {$c(-class) == "Checkbutton"} {
        
        if {$::DBG} {
            .t.c create rectangle $c(-ix) $c(-iy) $c(-ix1) $c(-iy1) \
                -fill white -outline $c(-darkshadow)
        }

        PdfRectangle $c(-ix) $c(-iy) \
                $c(-ix1) $c(-iy1) white $c(-darkshadow)
    
        PdfLine [list $c(-ix1) $c(-iy) $c(-ix1) $c(-iy1)] $c(-lightshadow) 
   
        PdfLine [list $c(-ix) $c(-iy1) $c(-ix1) $c(-iy1)] $c(-lightshadow)
        
        if {$::DBG} {
            .t.c create line $c(-ix1) $c(-iy) $c(-ix1) $c(-iy1)\
                -fill $c(-lightshadow)
        }

        incr c(-ix1)
         
        if {$::DBG} {
            .t.c create line $c(-ix) $c(-iy1) $c(-ix1) $c(-iy1) \
                -fill $c(-lightshadow)
        }
        
        # draw check
        if {$c(-selected)} {
            DrawCheckbuttonSelectMark "black" $c(-markcoords)
        }
    } else {
        
        if {$::DBG} {
            .t.c create oval $c(-ix) $c(-iy) $c(-ix1) $c(-iy1) \
                    -fill white -outline $c(-darkshadow)
        }

        PdfOval [list $c(-ix) $c(-iy) $c(-ix1) $c(-iy1)] \
                "white" $c(-darkshadow)

        # draw check
        if {$c(-selected)} {
            DrawRadiobuttonSelectMark "black" $c(-ix) $c(-iy) \
                    $c(-ix1) $c(-iy1)
        }
    }
}

#############################################################################
# proc DrawCheckbuttonSelectMark --
#
#       Draws select mark in checkbutton.
#
# Arguments:
#       color  - A color of select mark.
#       coords - A coordinates of check mark.
# Results:
#   The select mark in checkbutton will be drawn

proc DrawCheckbuttonSelectMark {color coords} {

    if {$::tktocanvas::scaleX != 1 || $::tktocanvas::scaleY != 1} { 
        set scaledCoords {}
        foreach {x y} $coords {
            set scaledCoords "$scaledCoords [ScaleCoords 0 $x $y]"
        }

        set coords $scaledCoords
    }
   
    if {$::tktocanvas::rotAngle != 0} {
        set rotatedCoords {}
 
        foreach {x y} $coords {
            set rotatedCoords "$rotatedCoords [::tktocanvas::rotateXY $x $y]"
        }

        set coords $rotatedCoords
    }

    if {($::tktocanvas::offsetX != 0) || ($::tktocanvas::offsetY != 0)} {
        set translatedCoords {} 
        foreach {x y} $coords {
            lappend translatedCoords [expr {$x + $::tktocanvas::offsetX}]
            lappend translatedCoords [expr {$y + $::tktocanvas::offsetY}]
        }

        set coords $translatedCoords
    }

    if {$::DBG} {
        .t.c create polygon $coords -outline $color -fill $color
    }

    PdfPolygon $coords $color $color 
}

#############################################################################
# proc DrawRadiobuttonSelectMark --
#
#       Draws select mark in checkbutton.
#
# Arguments:
#       color - A color of select mark.
#       ix    - X coordinates of top left edge of indicator's border.
#       iy    - Y coordinates of top left edge of indicator's border.
#       ix1   - X coordinates of bottom left edge of select mark.
#       iy1   - Y coordinates of bottom top edge of select mark.
# Results:
#   The select mark in radiobutton will be drawn.

proc DrawRadiobuttonSelectMark {color ix iy ix1 iy1} {

    incr ix 3
    incr iy 3
	set ix1 [expr $ix1 - 3]
    set iy1 [expr $iy1 - 3]
   
    if {$::DBG} {
        .t.c create oval $ix $iy $ix1 $iy1 -fill $color -outline $color
    }

    PdfOval [list $ix $iy $ix1 $iy1] $color $color
}

#############################################################################
# proc ListboxDrawElementsBg --
#
#       Draws a background for elements in listbox.
#
# Arguments:
#   config - a dictionary with configuration values.
# Results:
#   The background for each visible element in listbox will be drawn.
#
# Elementsinfo dictionary key is a list. Particular keys in that list:
# 0 - if element isn't selected the value will be -1
# 1 - text to be drawn
# 2 - text coords
# 3 - background and foreground colors for the element
# 4 - element's coordinates
#
# Visibleedges dictionary key possible values:
# 1 means only right edge is visible.
# 2 means only left ednge is visible.
# 3 means both edges are visible

proc ListboxDrawElementsBg {config} {

    array set c [dict get $config]
    
    set info [dict get $config -elementsinfo]
    
    set listLength [llength $info]
  
    for {set i 0} {$i < $listLength} {incr i} {
        set currEl [lindex $info $i]
        set textCoords [lindex $currEl 2]
        set colors [lindex $currEl 3]
        
        set isSel [lsearch $c(-selelements) [lindex $currEl 0]]
        
        foreach {bx by bx1 by1} [lindex $currEl 4] break

        TransformPoints bx by bx1 by1
        
        if {$::DBG} {
            .t.c create rectangle $bx $by $bx1 $by1 \
                        -fill [lindex $colors 1] -outline [lindex $colors 1]
        }

        PdfRectangle $bx $by $bx1 $by1 [lindex $colors 1] [lindex $colors 1]

        set textCoords [eval TkToPdfTextCoords $textCoords \
                [lindex $currEl 5] [list [dict get $config -font]]]
       
        
        if {$::DBG} {
            .t.c create text [lindex $textCoords 0] [lindex $textCoords 1] \
                    -text [lindex $currEl 1] -font [dict get $config -font] \
                    -fill [lindex $colors 0]
        }

        PdfText [lindex $textCoords 0] \
            [lindex $textCoords 1] \
            [dict get $config -font] \
            [lindex $colors 0] \
            [lindex $currEl 1] "left" \
            [lindex [lindex $currEl 5] 0] \
            [lindex [lindex $currEl 5] 1]

        if {$isSel >= 0} {
            set vShadowColors [lindex $colors 2]
            set hShadowColors [lindex $colors 2]
        
            set isVert1Visible 0
            set isVert2Visible 0
            set isHoriz1Visible 0
            set isHoriz2Visible 0
            set vertSide1 1
            set vertSide2 -1
            set horizSide1 1
            set horizSide2 -1
            set visibleEdges $c(-visibleedges)

            switch $::tktocanvas::rotAngle {
                0 - 180 {
                    if {[expr {$c(-visibleedges) & 1}]} {
                        set isVert1Visible 1
                    }
                    if {[expr {$c(-visibleedges) & 2}]} {
                        set isVert2Visible 1
                    }
                    if {!(([lindex $c(-selelements) [expr {$isSel - 1}]] \
                            == [expr {[lindex $currEl 0] - 1}]) \
                            && ($i != [lindex $c(-selelements) 0]))} {
                        set isHoriz1Visible 1
                    }
                    if {!([lindex $c(-selelements) [expr {$isSel + 1}]] \
                            == [expr {[lindex $currEl 0] + 1}])} {
                        set isHoriz2Visible 1
                    }

                    if {$::tktocanvas::rotAngle == 180} {
                        foreach {dark light} $vShadowColors {
                            lset vShadowColors 0 $light
                            lset vShadowColors 1 $dark
                        }
                        foreach {dark light} $hShadowColors {
                            lset hShadowColors 0 $light
                            lset hShadowColors 1 $dark
                        }
                        set vertSide1 -1
                        set vertSide2 1
                        set horizSide1 -1
                        set horizSide2 1
                        if {$visibleEdges == 1} {
                            set visibleEdges 2
                        } elseif {$visibleEdges == 2} {
                            set visibleEdges 1
                        }
                    }
                }
                default {
                    if {!(([lindex $c(-selelements) [expr {$isSel - 1}]] \
                            == [expr {[lindex $currEl 0] - 1}]) \
                            && ($i != [lindex $c(-selelements) 0]))} {
                        set isVert1Visible 1
                    }
                    if {!([lindex $c(-selelements) [expr {$isSel + 1}]] \
                            == [expr {[lindex $currEl 0] + 1}])} {
                        set isVert2Visible 1
                    }
                    if {[expr {$c(-visibleedges) & 1}]} {
                        set isHoriz1Visible 1
                    }
                    if {[expr {$c(-visibleedges) & 2}]} {
                        set isHoriz2Visible 1
                    }
                    if {$::tktocanvas::rotAngle == 90} {
                        foreach {dark light} $vShadowColors {
                            lset vShadowColors 0 $light
                            lset vShadowColors 1 $dark
                        }
                        set horizSide1 1
                        set horizSide2 -1
                        set vertSide1 -1
                        set vertSide2 1
                    } else {
                        foreach {dark light} $hShadowColors {
                            lset hShadowColors 0 $light
                            lset hShadowColors 1 $dark
                        }
                        set horizSide1 -1
                        set horizSide2 1
                        set vertSide1 1
                        set vertSide2 -1
                    }
                    
                    set prevEl [lindex $c(-selelements) [expr {$isSel - 1}]]
                    set nextEl [lindex $c(-selelements) [expr {$isSel + 1}]]
                    set visibleEdges 3
                    set isSel [lindex $c(-selelements) $isSel]
                    
                    if {($prevEl == [expr $isSel - 1]) \
                            && ($nextEl == [expr $isSel +1])} {
                        set visibleEdges 0
                    } elseif {$prevEl == [expr $isSel - 1]} {
                        if {$::tktocanvas::rotAngle == 90} {
                            set visibleEdges 1
                        } else {
                            set visibleEdges 2
                        }
                    } elseif {$nextEl == [expr $isSel + 1]} {
                        if {$::tktocanvas::rotAngle == 90} {
                            set visibleEdges 2
                        } else {
                            set visibleEdges 1
                        }
                    }
                }
            }
           
            if {$isVert1Visible} {
                DrawVerticalBar $bx $by $bx1 $by1 \
                        $c(-selectborderwidth) $vertSide1 $vShadowColors \
                        $c(-relief)
            }
            if {$isVert2Visible} {
                DrawVerticalBar $bx $by $bx1 $by1 \
                        $c(-selectborderwidth) $vertSide2 $vShadowColors \
                        $c(-relief)
            }
            if {$isHoriz1Visible} {
                DrawHorizontalBar $bx $by $bx1 $by1 \
                        $c(-selectborderwidth) $horizSide1 $hShadowColors \
                        $c(-relief) $visibleEdges                        
            }
            if {$isHoriz2Visible} {
                DrawHorizontalBar $bx $by $bx1 $by1 \
                        $c(-selectborderwidth) $horizSide2 $hShadowColors \
                        $c(-relief) $visibleEdges
            }
        }

    }
}

proc DrawScrollArrow {coords iCoords scolors bcolor width side} {
    
    if {[string range [lindex $scolors 0] 1 end] \
            > [string range [lindex $scolors 1] 1 end]} {
        
        set tmp [lindex $scolors 0]
        lset scolors 0 [lindex $scolors 1]
        lset scolors 1 $tmp
    }

    if {$side == 2} {
        set tmp [lindex $scolors 0]
        lset scolors 0 [lindex $scolors 1]
        lset scolors 1 $tmp
    }

    if {$::DBG} {
        .t.c create polygon $coords -fill [lindex $scolors 0]
    
        .t.c create polygon $iCoords -fill $bcolor

        .t.c create polygon [lindex $coords 0] [lindex $coords 1] \
                [lindex $iCoords 0] [lindex $iCoords 1] \
                [lindex $iCoords 2] [lindex $iCoords 3] \
                [lindex $coords 2] [lindex $coords 3] -fill [lindex $scolors 1]
    }

    PdfPolygon $iCoords $bcolor $bcolor
    PdfPolygon $coords [lindex $scolors 0] [lindex $scolors 0]
    PdfPolygon [list [lindex $coords 0] [lindex $coords 1] \
            [lindex $iCoords 0] [lindex $iCoords 1] \
            [lindex $iCoords 2] [lindex $iCoords 3] \
            [lindex $coords 2] [lindex $coords 3]] \
            [lindex $scolors 1] [lindex $scolors 1]
}

proc drawWindow {config} {

    puts "Name of exported window: [dict get $config -name]"

    set x [expr {[dict get $config -x] \
            - [dict get $config -highlightthickness]}]
    set y [expr {[dict get $config -y] \
            - [dict get $config -highlightthickness]}]
    set x1 [expr {[dict get $config -x1] \
            + [dict get $config -highlightthickness]}]
    set y1 [expr {[dict get $config -y1] \
            + [dict get $config -highlightthickness]}]

    TransformPoints x y x1 y1 

    SetDrawableRegion $x $y $x1 $y1
    
    set cmd "tktocanvas::draw[dict get $config -class]"
    eval $cmd {$config}

    if {[dict exists $config -child]} {
        foreach children [dict get $config -child] {
            drawWindow $children
        }
    }

    UnsetRegion
}

#######################################################################
# Procedure: drawButton - draw a button widget
# 
# Inputs:
# config - dictionary with config variables
#

proc drawButton {config} {

    # Do rotation and scaling.
    
    TransformDictPoints config

    # this a "body" of button	
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    PdfRectangle [dict get $config -x] [dict get $config -y] \
            [dict get $config -x1] [dict get $config -y1] \
            [dict get $config -background] [dict get $config -background]
    
    if {[dict get $config -haveimage]} {
       
        PdfImage {*}[dict get $config -imagecoords] \
                [dict get $config -imagename] [dict get $config -imagewidth] \
                [dict get $config -imageheight] [dict get $config -data]
        
        # Tune image which will be drawn to size from dictionary.
        
        set img [image create photo -file [[dict get $config -image] cget -file]]
        $img configure -width [dict get $config -imagewidth]  
        $img configure -height [dict get $config -imageheight] 
        dict set config -image $img 

        if {$::DBG} {
            .t.c create image [dict get $config -imagecoords] -image \
                    [dict get $config -image] -anchor nw
        }

    }
    
    if {[dict get $config -havetext]} {
        
        dict set config -textcoords [TkToPdfTextCoords \
                [lindex [dict get $config -textcoords] 0] \
                [lindex [dict get $config -textcoords] 1] \
                [dict get $config -textwidth] \
                [dict get $config -textheight] \
                [dict get $config -font]]
        
        if {$::DBG} {
            .t.c create text [dict get $config -textcoords] \
                    -text [dict get $config -text] -font [dict get $config -font] \
                    -justify [dict get $config -justify] \
                    -underline [dict get $config -underline]
        }

        PdfText [lindex [dict get $config -textcoords] 0] \
            [lindex [dict get $config -textcoords] 1]\
            [dict get $config -font] \
            [dict get $config -foreground] \
            [dict get $config -text] [dict get $config -justify] \
            [dict get $config -textwidth] \
            [dict get $config -textheight]
    }

    if {[dict get $config -relief] != "flat"} {
    
        Draw3DBorder [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -relief] [dict get $config -borderwidth] \
                [list [dict get $config -lightshadow] \
                [dict get $config -darkshadow]]
    }
    
    DrawHighlightthickness $config
}

proc drawCheckbutton {config} {
	
    drawButton $config
   
    TransformDictPoints config ix iy ix1 iy1
    
    if {[dict get $config -indicatoron]} {
        BtnDrawIndicator $config
    }
}

proc drawRadiobutton {config} {

    drawButton $config

    TransformDictPoints config ix iy ix1 iy1
    
    if {[dict get $config -indicatoron]} {
        BtnDrawIndicator $config
    }
}

proc drawLabel {config} {

    drawButton $config
}

proc drawFrame {config} {

    # If frame has container option seted to true, we do not 
    # draw this frame.

    if {$config == 0} {
        return
    }

    TransformDictPoints config
    
    if {!([dict get $config -background] == "")} {
        
        if {$::DBG} {
            .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                    [dict get $config -x1] [dict get $config -y1] \
                    -fill [dict get $config -background] \
                    -outline [dict get $config -background]
        }

        PdfRectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -background] \
                [dict get $config -background]
   
        if {[dict get $config -class] == "Labelframe"} {
            
            foreach {sx sy sx1 sy1} "[dict get $config -shiftx] \
                    [dict get $config -shifty] [dict get $config -shiftx1] \
                    [dict get $config -shifty1]" break
            
            switch $::tktocanvas::rotAngle {
                90 {
                    if {$sx} {
                        set sy $sx
                        set sx 0
                    } elseif {$sy} {
                        set sx1 -$sy
                        set sy 0
                    } elseif {$sx1} {
                        set sy1 -$sx1
                        set sx1 0
                    } elseif {$sy1} {
                        set sx $sy1
                        set sy1 0
                    }
                }
                180 {
                    if {$sx} {
                        set sx1 -$sx
                        set sx 0
                    } elseif {$sy} {
                        set sy1 -$sy
                        set sy 0
                    } elseif {$sx1} {
                        set sx $sx1
                        set sx1 0
                    } elseif {$sy1} {
                        set sy $sy1
                        set sy1 0
                    }
                }
                270 {
                    if {$sx} {
                        set sy1 -$sx
                        set sx 0
                    } elseif {$sy} {
                        set sx $sy
                        set sy 0
                    } elseif {$sx1} {
                        set sy $sx1
                        set sx1 0
                    } elseif {$sy1} {
                        set sx1 -$sy1
                        set sy1 0
                    }
                }
            }

            dict set config -shiftx [expr {int($sx * $::tktocanvas::scaleX)}]
            dict set config -shifty [expr {int($sy * $::tktocanvas::scaleY)}]
            dict set config -shiftx1 [expr {int($sx1 * $::tktocanvas::scaleX)}]
            dict set config -shifty1 [expr {int($sy1 * $::tktocanvas::scaleY)}]

            dict set config -x [expr {[dict get $config -x] \
                    + [dict get $config -shiftx]}]
            dict set config -y [expr {[dict get $config -y] \
                    + [dict get $config -shifty]}]
            dict set config -x1 [expr {[dict get $config -x1] \
                    + [dict get $config -shiftx1]}]
            dict set config -y1 [expr {[dict get $config -y1] \
                    + [dict get $config -shifty1]}]
        } else {
            DrawHighlightthickness $config
        }

        if {[dict get $config -relief] != "flat"} {
            Draw3DBorder [dict get $config -x] [dict get $config -y] \
                    [dict get $config -x1] [dict get $config -y1] \
                    [dict get $config -relief] [dict get $config -borderwidth]\
                    [list [dict get $config -lightshadow] \
                    [dict get $config -darkshadow]]
        }

    } else {
        # If it has -background color set up to "" then
        # only highlightthickness is drawn.

        # TODO: write function to get background color from parent window

        if {$::DBG} {
            .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                    [dict get $config -x1] [dict get $config -y1] \
                    -outline [. cget -bg] -fill [. cget -bg]
        }

        PdfRectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -background] \
                [dict get $config -background]
    }
    
}

proc drawLabelframe {config} {

    # If frame has container option seted to true, we do not 
    # draw this frame.
    
    if {$config == 0} {
        return
    }
    
    drawFrame $config

    TransformDictPoints config
    
    if {[dict get $config -labelwidget] == ""} {

        dict set config -labelboxcoords [TransformPointsList \
                [dict get $config -labelboxcoords]]
        
        dict set config -textcoords [TkToPdfTextCoords \
                [lindex [dict get $config -textcoords] 0] \
                [lindex [dict get $config -textcoords] 1] \
                [dict get $config -textwidth] \
                [dict get $config -textheight] \
                [dict get $config -font]]
        
        PdfRectangle {*}[dict get $config -labelboxcoords] \
                [dict get $config -background] [dict get $config -background]
        
        PdfText [lindex [dict get $config -textcoords] 0] \
                [lindex [dict get $config -textcoords] 1] \
                [dict get $config -font] \
                [dict get $config -foreground] \
                [dict get $config -text] "center"  \
                [dict get $config -textwidth] \
                [dict get $config -textheight]
       
        if {$::DBG} {
            .t.c create rectangle [dict get $config -labelboxcoords] \
                    -fill [dict get $config -background] \
                    -outline [dict get $config -background]
            .t.c create text [dict get $config -textcoords] \
                    -text [dict get $config -text] -font [dict get $config -font] \
                    -fill [dict get $config -foreground]
        }
    }
    
    DrawHighlightthickness $config
}

proc drawEntry {config} {

    TransformDictPoints config
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    PdfRectangle [dict get $config -x] [dict get $config -y] \
            [dict get $config -x1] [dict get $config -y1] \
            [dict get $config -background] \
            [dict get $config -background]
     
    dict set config -textcoords [TkToPdfTextCoords \
        [lindex [dict get $config -textcoords] 0] \
        [lindex [dict get $config -textcoords] 1] \
        [dict get $config -textwidth] \
        [dict get $config -textheight] \
        [dict get $config -font]]
    
    #dict set config -textcoords [TransformPointsList \
        [dict get $config -textcoords]]
    
    PdfText {*}[dict get $config -textcoords] \
        [dict get $config -font] \
        "black" \
        [dict get $config -text] "left"  \
        [dict get $config -textwidth] \
        [dict get $config -textheight]
 
    if {$::DBG} {
        .t.c create text [dict get $config -textcoords] \
                -text [dict get $config -text] \
                -font [dict get $config -font] \
                -justify [dict get $config -justify]
    }

    if {[dict get $config -relief] != "flat"} {
        Draw3DBorder [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -relief] [dict get $config -borderwidth]\
                [list [dict get $config -lightshadow] \
                [dict get $config -darkshadow]]
    }
    
    DrawHighlightthickness $config
}

proc drawSpinbox {config} {
    
    drawEntry $config

    TransformDictPoints config bupx bupy bupx1 bupy1
    
    PdfRectangle [dict get $config -bupx] [dict get $config -bupy] \
            [dict get $config -bupx1] [dict get $config -bupy1] \
            [dict get $config -background] \
            [dict get $config -background]

    if {$::DBG} {
        .t.c create rectangle [dict get $config -bupx] [dict get $config -bupy] \
                [dict get $config -bupx1] [dict get $config -bupy1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    Draw3DBorder [dict get $config -bupx] [dict get $config -bupy] \
            [dict get $config -bupx1] [dict get $config -bupy1] \
            [dict get $config -buttonuprelief] 1 \
            [dict get $config -upbuttonshadows]

    TransformDictPoints config bdownx bdowny bdownx1 bdowny1
    
    PdfRectangle [dict get $config -bdownx] \
            [dict get $config -bdowny] \
            [dict get $config -bdownx1] [dict get $config -bdowny1] \
            [dict get $config -background] \
            [dict get $config -background]
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -bdownx] \
                [dict get $config -bdowny] \
                [dict get $config -bdownx1] [dict get $config -bdowny1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
        
        .t.c create polygon {*}[dict get $config -upButtonCoords] \
                -fill black -outline black
    }

    Draw3DBorder [dict get $config -bdownx] [dict get $config -bdowny] \
            [dict get $config -bdownx1] [dict get $config -bdowny1] \
            [dict get $config -buttondownrelief] 1 \
            [dict get $config -downbuttonshadows]

    dict set config -upButtonCoords [TransformPointsList \
            [dict get $config -upButtonCoords]]
    PdfPolygon [dict get $config -upButtonCoords] "black" ""
    
    if {$::DBG} {
        .t.c create polygon {*}[dict get $config -downButtonCoords] \
                -fill black -outline black
    }

    dict set config -downButtonCoords [TransformPointsList \
            [dict get $config -downButtonCoords]]
    PdfPolygon [dict get $config -downButtonCoords] "black" ""
}

proc drawListbox {config} {

    # Do rotation and scaling.
    
    TransformDictPoints config
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

     #SetDrawableRegion [expr {[dict get $config -x] \
            + [dict get $config -borderwidth]}] \
            [expr {[dict get $config -y] + [dict get $config -borderwidth]}] \
            [expr {[dict get $config -x1] - [dict get $config -borderwidth]}]\
            [expr {[dict get $config -y1] - [dict get $config -borderwidth]}]\
    
    ListboxDrawElementsBg $config

    #UnsetRegion

    if {[dict get $config -relief] != "flat"} {
        Draw3DBorder [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -relief] [dict get $config -borderwidth]\
                [list [dict get $config -lightshadow] \
                [dict get $config -darkshadow]]
    }
    
    DrawHighlightthickness $config
}

proc drawMenubutton {config} {
    
    drawButton $config
}

proc drawMenu {config} {

    TransformDictPoints config
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    PdfRectangle [dict get $config -x] [dict get $config -y] \
            [dict get $config -x1] [dict get $config -y1] \
            [dict get $config -background] \
            [dict get $config -background]

    foreach entry [dict get $config -entriesinfo] {
       
        # Draw background for each entry
        
        TransformDictPoints entry
        
        if {$::DBG} {
            .t.c create rectangle [dict get $entry -x] [dict get $entry -y] \
                    [dict get $entry -x1] [dict get $entry -y1] \
                    -fill [dict get $entry -bg] -outline [dict get $entry -bg]
        }

        PdfRectangle [dict get $entry -x] [dict get $entry -y] \
                [dict get $entry -x1] [dict get $entry -y1] \
                [dict get $entry -bg] [dict get $entry -bg]

        if {[dict get $entry -image] != ""} {
            
            if {$::DBG} {
                .t.c create image [dict get $entry -imagecoords] \
                        -image [dict get $entry -image] -anchor nw
            }

            PdfImage {*}[dict get $entry -imagecoords] \
                    [dict get $entry -imagename] [dict get $entry -imagewidth] \
                    [dict get $entry -imageheight] [dict get $entry -data]
        }
        
        dict set entry -textcoords [eval TkToPdfTextCoords [dict get $entry -textcoords] \
                [dict get $entry -textwidth] [dict get $entry -textheight] \
                [list [dict get $entry -font]]]
        #dict set entry -textcoords [TransformPointsList \
                [dict get $entry -textcoords]]
        
        if {$::DBG} {
            .t.c create text [dict get $entry -textcoords] \
                    -text [dict get $entry -text] -fill [dict get $entry -fg] \
                    -font [dict get $entry -font]
        }

        PdfText {*}[dict get $entry -textcoords] [dict get $entry -font] \
                [dict get $entry -foreground] [dict get $entry -text] \
                "left" [dict get $entry -textwidth] [dict get $entry -textheight]
        
        if {[dict get $entry -type] == "checkbutton" \
                && [dict get $entry -selected] != 0} {
           
            DrawCheckbuttonSelectMark [dict get $entry -selectcolor] \
                    [dict get $entry -markcoords]
        }
        if {[dict get $entry -type] == "radiobutton" \
                && [dict get $entry -selected] != 0} {
            
            TransformDictPoints entry ix iy ix1 iy1
            
            DrawRadiobuttonSelectMark [dict get $entry -selectcolor] \
                    [dict get $entry -ix] [dict get $entry -iy] \
                    [dict get $entry -ix1] [dict get $entry -iy1]
        }
    }

    if {[dict get $config -relief] != "flat"} {
        Draw3DBorder [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -relief] [dict get $config -borderwidth]\
                [list [dict get $config -lightshadow] \
                [dict get $config -darkshadow]]
    }
}

proc drawCanvas {config} {
    
    TransformDictPoints config
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    PdfRectangle [dict get $config -x] [dict get $config -y] \
            [dict get $config -x1] [dict get $config -y1] \
            [dict get $config -background] \
            [dict get $config -background]
  
    #SetDrawableRegion [dict get $config -x] [dict get $config -y] \
            [dict get $config -x1] [dict get $config -y1]

    foreach item [dict get $config -itemsinfo] {
        
        set class [dict get $item -class]
        set coords [dict get $item -coords]

        set coords [TransformPointsList $coords]

        SetStandardOptions item
       
        switch $class {
            "rectangle" {
                PdfRectangle {*}$coords [dict get $item -fill] \
                        [dict get $item -outline] \
                        [dict get $item -width]
            }
            "oval" {
                PdfOval $coords [dict get $item -fill] \
                        [dict get $item -outline] [dict get $item -width]
            }
            "polygon" {
                PdfPolygon $coords [dict get $item -fill] \
                        [dict get $item -outline] [dict get $item -width] \
                        [dict get $item -smooth]
            }
            "arc" {
                if {$::tktocanvas::rotAngle != 0} {             
                    dict set item -start [expr {[dict get $item -start] \
                            + $::tktocanvas::rotAngle}]
                }

                PdfArc {*}$coords [dict get $item -fill] \
                        [dict get $item -outline] [dict get $item -start] \
                        [dict get $item -extent] [dict get $item -style] \
                        [dict get $item -width]
            }
            "line" {
                PdfLine $coords [dict get $item -fill] \
                        [dict get $item -width] [dict get $item -smooth] \
                        [dict get $item -arrow] [dict get $item -arrowshape]
            }
            "image" {
                PdfImage {*}$coords [dict get $item -imagename] \
                        [dict get $item -width] [dict get $item -height] \
                        [dict get $item -data]

                        #set i [image create photo -data [dict get $item -ddd] -width 50 -height 50]
                        #.t.c create image 50 50 -image $i
            }
            "text" {
                set coords [TkToPdfTextCoords {*}$coords [dict get $item -textwidth] \
                        [dict get $item -textheight] [dict get $item -font] \
                        [dict get $item -anchor]]
                PdfText {*}$coords [dict get $item -font] [dict get $item -fill] \
                        [dict get $item -text] [dict get $item -justify] \
                        [dict get $item -textwidth] [dict get $item -textheight]
            }
        }

        set itemOptions [dict remove $item -coords -class]
        if {$::DBG} {
            set cmd ".t.c create $class $coords $itemOptions"
            
            if {$class != "image" && $class != "text"} { 
                eval $cmd
            }
        }
    }

    #UnsetRegion
    
    if {[dict get $config -relief] != "flat"} {
        Draw3DBorder [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -relief] [dict get $config -borderwidth]\
                [list [dict get $config -lightshadow] \
                [dict get $config -darkshadow]]
    }

    DrawHighlightthickness $config
}

proc drawScale {config} {
    
    TransformDictPoints config
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    PdfRectangle [dict get $config -x] [dict get $config -y] \
            [dict get $config -x1] [dict get $config -y1] \
            [dict get $config -background] [dict get $config -background]
    
    TransformDictPoints config troughx troughy troughx1 troughy1
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -troughx] \
                [dict get $config -troughy] [dict get $config -troughx1] \
                [dict get $config -troughy1] \
                -fill [dict get $config -troughcolor] \
                -outline [dict get $config -troughcolor]
    }

    PdfRectangle [dict get $config -troughx] \
            [dict get $config -troughy] [dict get $config -troughx1] \
            [dict get $config -troughy1] \
            [dict get $config -troughcolor] \
            [dict get $config -troughcolor]
    
    if {[dict get $config -relief] != "flat"} {
        Draw3DBorder [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -relief] [dict get $config -borderwidth]\
                [list [dict get $config -lightshadow] \
                [dict get $config -darkshadow]]
    }

    Draw3DBorder [dict get $config -troughx] [dict get $config -troughy] \
            [dict get $config -troughx1] [dict get $config -troughy1] \
            "sunken" [dict get $config -borderwidth]\
            [list [dict get $config -troughlightshadow] \
            [dict get $config -troughdarkshadow]]
    
    TransformDictPoints config sliderx slidery sliderx1 slidery1
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -sliderx] [dict get $config -slidery] \
                [dict get $config -sliderx1] [dict get $config -slidery1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    PdfRectangle [dict get $config -sliderx] [dict get $config -slidery] \
            [dict get $config -sliderx1] [dict get $config -slidery1] \
            [dict get $config -background] \
            [dict get $config -background]

    Draw3DBorder [dict get $config -sliderx] [dict get $config -slidery] \
            [dict get $config -sliderx1] [dict get $config -slidery1] \
            [dict get $config -sliderrelief] \
            [expr {([dict get $config -borderwidth]+1) / 2}] \
            [list [dict get $config -sliderlightshadow] \
            [dict get $config -sliderdarkshadow]]
    
    TransformDictPoints config spartx sparty spartx1 sparty1
    
    Draw3DBorder [dict get $config -spartx] [dict get $config -sparty] \
            [dict get $config -spartx1] [dict get $config -sparty1] \
            [dict get $config -sliderrelief] \
            [expr {([dict get $config -borderwidth]+1) / 2}] \
            [list [dict get $config -sliderlightshadow] \
            [dict get $config -sliderdarkshadow]]
    
    TransformDictPoints config spartpx spartpy spartpx1 spartpy1
    
    Draw3DBorder [dict get $config -spartpx] [dict get $config -spartpy] \
            [dict get $config -spartpx1] [dict get $config -spartpy1] \
            [dict get $config -sliderrelief] \
            [expr {([dict get $config -borderwidth]+1) / 2}] \
            [list [dict get $config -sliderlightshadow] \
            [dict get $config -sliderdarkshadow]]
    
    foreach Tick [dict get $config -tick] {
        lset Tick 0 [TkToPdfTextCoords {*}[lindex $Tick 0] \
                [lindex $Tick 2] \
                [lindex $Tick 3] \
                [dict get $config -font]]
        #lset Tick 0 [TransformPointsList [lindex $Tick 0]]
        
        PdfText {*}[lindex $Tick 0] [dict get $config -font] \
                [dict get $config -foreground] [lindex $Tick 1] \
                "left" [lindex $Tick 2] [lindex $Tick 3]
        
        if {$::DBG} {
            .t.c create text [lindex $Tick 0] -text [lindex $Tick 1] \
                -font [dict get $config -font] -fill [dict get $config -foreground]
        }
    }
    
    if {[dict get $config -showvalue]} {
        
        dict set config -valuecoords [TkToPdfTextCoords \
                {*}[dict get $config -valuecoords] \
                {*}[dict get $config -valuetextlayout] \
                [dict get $config -font]]
        
        #dict set config -valuecoords \
                [TransformPointsList [dict get $config -valuecoords]]
        
        PdfText {*}[dict get $config -valuecoords] \
                [dict get $config -font] [dict get $config -foreground] \
                [dict get $config -value] "left" \
                {*}[dict get $config -valuetextlayout]]
        
        if {$::DBG} {
            .t.c create text [dict get $config -valuecoords] \
                    -text [dict get $config -value] -font [dict get $config -font]\
                    -fill [dict get $config -foreground]
        }
    } 
    
    if {[dict get $config -label] != ""} {
        dict set config -labelcoords [TkToPdfTextCoords \
                {*}[dict get $config -labelcoords] \
                {*}[dict get $config -labeltextlayout] \
                [dict get $config -font]]
        
        #dict set config -labelcoords \
                [TransformPointsList [dict get $config -labelcoords]]
        
        PdfText {*}[dict get $config -labelcoords] \
                [dict get $config -font] [dict get $config -foreground] \
                [dict get $config -label] "left" \
                {*}[dict get $config -labeltextlayout]]
        
        if {$::DBG} {
            .t.c create text [dict get $config -labelcoords] \
                    -text [dict get $config -label] -font [dict get $config -font]\
                    -fill [dict get $config -foreground]
        }
    } 

    DrawHighlightthickness $config
}

proc drawScrollbar {config} {

    TransformDictPoints config
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    PdfRectangle [dict get $config -x] [dict get $config -y] \
            [dict get $config -x1] [dict get $config -y1] \
            [dict get $config -background] \
            [dict get $config -background]

    dict set config -arrowcoords [TransformPointsList \
            [dict get $config -arrowcoords]]
    dict set config -interiorcoords [TransformPointsList \
            [dict get $config -interiorcoords]]

    DrawScrollArrow [dict get $config -arrowcoords] \
            [dict get $config -interiorcoords] \
            [dict get $config -elementsshadows] [dict get $config -background] \
            [dict get $config -elementborderwidth] 1
   
    dict set config -arrowcoords1 [TransformPointsList \
            [dict get $config -arrowcoords1]]
    dict set config -interiorcoords1 [TransformPointsList \
            [dict get $config -interiorcoords1]]
    
    DrawScrollArrow [dict get $config -arrowcoords1] \
            [dict get $config -interiorcoords1] \
            [dict get $config -elementsshadows] [dict get $config -background] \
            [dict get $config -elementborderwidth] 2

    TransformDictPoints config sliderx slidery sliderx1 slidery1
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -sliderx] [dict get $config -slidery] \
                [dict get $config -sliderx1] [dict get $config -slidery1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    PdfRectangle [dict get $config -sliderx] [dict get $config -slidery] \
            [dict get $config -sliderx1] [dict get $config -slidery1] \
            [dict get $config -background] \
            [dict get $config -background]
    
    Draw3DBorder [dict get $config -sliderx] [dict get $config -slidery] \
            [dict get $config -sliderx1] [dict get $config -slidery1] \
            "raised" [dict get $config -sliderborderwidth] \
            [dict get $config -elementsshadows] 

    if {[dict get $config -relief] != "flat"} {
        Draw3DBorder [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -relief] [dict get $config -borderwidth]\
                [list [dict get $config -lightshadow] \
                [dict get $config -darkshadow]]
    }

    DrawHighlightthickness $config
}

proc drawToplevel {config} {
    
    if {$::DBG} {
        .t.c create rectangle [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                -fill [dict get $config -background] \
                -outline [dict get $config -background]
    }

    TransformDictPoints config
    
    PdfRectangle [dict get $config -x] [dict get $config -y] \
            [dict get $config -x1] [dict get $config -y1] \
            [dict get $config -background] \
            [dict get $config -background]

    if {[dict get $config -relief] != "flat"} {
        Draw3DBorder [dict get $config -x] [dict get $config -y] \
                [dict get $config -x1] [dict get $config -y1] \
                [dict get $config -relief] [dict get $config -borderwidth]\
                [list [dict get $config -lightshadow] \
                [dict get $config -darkshadow]]
    }
    
    DrawHighlightthickness $config
}

proc drawPanedwindow {config} {
    
    drawToplevel $config
}

proc drawText {config} {

    drawToplevel $config
}

proc prepareToDraw {} {
    
    wm geometry . =400x600+100+50
   
    update idletasks
   
    set Geom [wm geometry .]
    set XSize [string first "x" $Geom]
    set XSize [string range $Geom 0 $XSize-1]

    set XPos [string range $Geom \
        [string first "+" $Geom]+1 [string last "+" $Geom]-1]
    incr XPos $XSize
    incr XPos 6

    set Geom [string replace $Geom [string first "+" $Geom]+1 \
        [string last "+" $Geom]-1 $XPos]
    
    toplevel .t
    wm geometry .t $Geom
    
    bind . <q> exit
    bind .t <q> exit
    
    canvas .t.c -width 1024 -height 768 -bg [. cget -bg] -borderwidth 0 \
        -highlightthickness 0
    place .t.c -x 0 -y 0

    update idletasks
}

} ;# end of namespace tktocanvas


